/***************************************************************************
                           chromakill.cpp
                           --------------
    begin                : Sat Jul 31 2004
    copyright            : (C) 2004 by Dirk Ziegelmeier
    email                : dziegel@gmx.de
 ***************************************************************************/

/*
 * ChromaKill algorithm taken from DScaler.
 * Copyright (c) 2003 Laurent Garnier. All rights reserved.
 *
 * Ported by: Dirk Ziegelmeier
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 */

#include <kdebug.h>
#include <klocale.h>
#include <kconfig.h>

#include "chromakill.h"
#include "kdetvimagefilter.h"
#include "kdetvimagefiltercontext.h"
#include "kdetvcpudetection.h"

// Qt 3.1.1 compatibility
#ifndef Q_INT64
#define Q_INT64 long long
#endif

class ChromaKillImageFilter : public KdetvImageFilter
{
public:
    ChromaKillImageFilter()
        : KdetvImageFilter("ChromaKill")
    {
    }

    virtual ~ChromaKillImageFilter()
    {
    }

    virtual KdetvImageFilterContext* operator<< (KdetvImageFilterContext* ctx)
    {
        // Return if we don't have MMX
        if(!(KdetvCpuDetection::capabilities() & KdetvCpuDetection::Cap_MMX)) {
            return ctx;
        }

        Q_INT64        qwLumiMask       = 0x00FF00FF00FF00FFull;
        Q_INT64        qwChromaHalfMask = 0x8000800080008000ull;
        int            Cycles           = ctx->out->bytesPerLine() / 8;
        unsigned char* Pixels           = ctx->out->buffer();
        int            fieldHeight      = ctx->out->size().height();
        unsigned int   outputpitch      = ctx->out->bytesPerLine() + ctx->out->stride();

        #define _qwLumiMask       "%0"
        #define _qwChromaHalfMask "%1"
        #define _Pixels           "%2"
        #define _Cycles           "%3"

        for (int y=0; y<fieldHeight; y++) {
            __asm__ __volatile__ (
                "movq    "_qwLumiMask",       %%mm1\n\t"
                "movq    "_qwChromaHalfMask", %%mm2\n\t"

                "1:\n\t"
                "movq    ("_Pixels"),         %%mm0\n\t"

                "pand    %%mm1,               %%mm0\n\t"
                "por     %%mm2,               %%mm0\n\t"

                "movq    %%mm0,               ("_Pixels")\n\t"

                "addl    $8,                  "_Pixels"\n\t"
                "decl    "_Cycles"\n\t"
                "jne     1b\n\t"

                : /* no outputs */

                : "m"(qwLumiMask),
                  "m"(qwChromaHalfMask),
                  "q"(Pixels),
                  "q"(Cycles)

                : "st", "st(1)",
                  "mm0", "mm1",
                  "memory", "cc"
            );

            Pixels += outputpitch;
        }

        __asm__ __volatile__ ("emms\n\t");

        return ctx;
    }

    virtual KdetvImage::ImageFormat inputFormats()
    {
        return KdetvImage::FORMAT_YUV422;
    };
};

// -----------------------------------------------------------------------

ChromaKillPlugin::ChromaKillPlugin(Kdetv *ktv, const QString& cfgkey, QObject *parent, const char* name)
    : KdetvFilterPlugin(ktv, cfgkey, parent, name)
{
    _filter = new ChromaKillImageFilter();
}

ChromaKillPlugin::~ChromaKillPlugin()
{
    delete _filter;
}

extern "C" {
    ChromaKillPlugin* create_chromakill(Kdetv* ktv)
    {
        return new ChromaKillPlugin(ktv, "chromakill", 0, "Kill image chroma");
    }
}

#include "chromakill.moc"
